﻿using System;
using System.Collections.Generic;
using System.Text;

namespace COMET2Lib
{
    public class Assembler
    {
        private Interpreter interp;

        // 5パターンに分けられた命令表
        private Dictionary<string, ushort> none = new Dictionary<string, ushort>()
        {
            {"NOP", 0x0 },
            {"RET", 0x8100 },
        };
        private Dictionary<string, ushort> r_adr_x = new Dictionary<string, ushort>()
        {
            { "LD", 0x1000 },
            { "ST", 0x1100 },
            { "LAD", 0x1200 },
            { "ADDA", 0x2000 },
            { "SUBA", 0x2100 },
            { "ADDL", 0x2200 },
            { "SUBL", 0x2300 },
            { "AND", 0x3000 },
            { "OR", 0x3100 },
            { "XOR", 0x3200 },
            { "CPA", 0x4000 },
            { "CPL", 0x4100 },
            { "SLA", 0x5000 },
            { "SRA", 0x5100 },
            { "SLL", 0x5200 },
            { "SRL", 0x5300 },
        };
        private Dictionary<string, ushort> r1_r2 = new Dictionary<string, ushort>()
        {
            { "LD", 0x1400 },
            { "ADDA", 0x2400 },
            { "SUBA", 0x2500 },
            { "ADDL", 0x2600 },
            { "SUBL", 0x2700 },
            { "AND", 0x3400 },
            { "OR", 0x3500 },
            { "XOR", 0x3600 },
            { "CPA", 0x4400 },
            { "CPL", 0x4500 },
        };
        private Dictionary<string, ushort> adr_x = new Dictionary<string, ushort>()
        {
            { "JMI", 0x6100 },
            { "JNZ", 0x6200 },
            { "JZE", 0x6300 },
            { "JUMP", 0x6400 },
            { "JPL", 0x6500 },
            { "JOV", 0x6600 },
            { "PUSH", 0x7000 },
            { "CALL", 0x8000 },
            { "SVC", 0xF000 },
        };
        private Dictionary<string, ushort> r = new Dictionary<string, ushort>()
        {
            { "POP", 0x7100 },
        };

        // アセンブラが生成したデータを書いていくポインタ
        private ushort bwp;	// binary writing pointer

        // 構文エラーの通知
        private bool syntaxError(string srcLine)
        {
            Console.Error.WriteLine("Syntax Error: {0}", srcLine);
            return false;
        }

        // ホワイトスペースの判定
        private bool iswhitespace(char ch) { return ch == ' ' || ch == '\t'; }

        // 文字列のレジスタ名からレジスタ番号を整数で抜き出す
        private int getRegisterNumber(string str)
        {
            if (str.Length != 3) return -1;
            if (str[0] != 'G' || str[1] != 'R') return -1;
            if (str[2] < '0' || str[2] > '7') return -1;
            return str[2] - '0';
        }

        // 1行の中の1トークンを解析して取り出す
        private string[] divideTokens(string src)
        {
            var list = new List<string>();
            bool quoting = false, skipping = false;
            StringBuilder sb = new StringBuilder();
            foreach (char ch in src)
            {
                // いい加減な実装だが、それでも'…''…'を適切にさばける
                // ''→'の処理は別の場所でやるので、ここで意識する必要はない
                if (ch == '\'') quoting = !quoting;
                if (ch == ';') break;
                if (skipping)
                {
                    skipping = iswhitespace(ch);
                    if (skipping) continue;
                }
                else if (!quoting && (iswhitespace(ch) || ch == ','))
                {
                    list.Add(sb.ToString());
                    sb.Length = 0;
                    skipping = true;
                    continue;
                }
                sb.Append(ch);
            }
            list.Add(sb.ToString());
            return list.ToArray();
        }

        // ラベルテーブル
        private Dictionary<string, ushort> labelTable = new Dictionary<string, ushort>();

        // 引数のr1を命令第1語に反映
        private bool generateR1(string str, ref ushort inst)
        {
            int n = getRegisterNumber(str);
            inst |= (ushort)(n << 4);
            return n >= 0;
        }

        // 引数のr2を命令第1語に反映
        private bool generateR2(string str, ref ushort inst)
        {
            int n = getRegisterNumber(str);
            inst |= (ushort)n;
            return n >= 0;
        }

        // 引数のxを命令第1語に反映
        // generateR2とは正当なレジスタの判定条件が違うことに注意
        // R2にGR0はあるが、XにGR0は無い
        private bool generateX(string str, ref ushort inst)
        {
            int n = getRegisterNumber(str);
            inst |= (ushort)n;
            return n > 0;
        }

        // 定数を解析する
        private bool constantParser(string str, out ushort target, int pass)
        {
            if (ushort.TryParse(str, out target)) return true;
            if (str[0] == '#' && ushort.TryParse(str.Substring(1), System.Globalization.NumberStyles.AllowHexSpecifier, null, out target)) return true;
            if (pass == 1 || labelTable.TryGetValue(str, out target)) return true;
            Console.Error.WriteLine("label {0} not found", str);
            return false;
        }

        // =で前置された即値を格納する場所を示す
        private ushort constReservePtr = 0xf000;

        // adrの値を生成する
        // パス1では0を入れているところと、
        // =で前置された即値の処理が違うことに注意
        private bool generateAdr(string str, out ushort adr, int pass)
        {
            if (!str.StartsWith("=")) return constantParser(str, out adr, pass);
            ushort target;
            adr = 0;
            if (!constantParser(str.Substring(1), out target, pass)) return false;
            if (pass == 1) return true;
            interp.memory[constReservePtr] = target;
            adr = constReservePtr++;
            return true;
        }

        // DC命令専用の値の生成
        private bool generateDC(string str, int pass)
        {
            ushort target;
            if (str.Length == 0) return true;
            if (str[0] == '\'')
            {
                str = str.Substring(1);
                while (str.Length != 0)
                {
                    if (str[0] == '\'')
                    {
                        if (str.Length < 2 || str[1] != '\'') break;
                        str = str.Substring(1);
                    }
                    interp.memory[bwp++] = str[0];
                    str = str.Substring(1);
                }
            }
            else
                if (!constantParser(str, out target, pass)) return false;
                else interp.memory[bwp++] = target;
            return true;
        }

        // 1行のアセンブル
        private bool assembleLine(string line, int pass, ref bool endFlag)
        {
            ushort inst, adr;

            // 行の解析結果を入れる変数
            string[] tokens = divideTokens(line);
            string label = tokens.Length >= 1 ? tokens[0] : string.Empty;
            string mnem = tokens.Length >= 2 ? tokens[1] : string.Empty;
            string[] args = new string[3];
            for (int i = 0; i < 3; i++)
                args[i] = tokens.Length >= i + 3 ? tokens[i + 2] : string.Empty;

            if (pass == 1 && label.Length > 0)
            {
                if (labelTable.ContainsKey(label))
                {
                    Console.Error.WriteLine("label {0} already defined", label);
                    return false;
                }
                labelTable[label] = bwp;
            }
            if (mnem.Length == 0)
            { }	// empty mnem
            else if (mnem == "START")
                interp.PR = bwp;
            else if (mnem == "END")
                endFlag = true;
            else if (mnem == "DS")
                bwp += (ushort)int.Parse(args[0]);	// エラーチェック手抜き
            else if (mnem == "DC")
                for (int i = 2; i < tokens.Length; i++)
                    if (!generateDC(tokens[i], pass)) return false; else { }
            else if (mnem == "IN")
            {
                interp.memory[bwp++] = 0x7001;	//PUSH GR1
                interp.memory[bwp++] = 0;
                interp.memory[bwp++] = 0x1210;	//LAD GR1,args[0]
                if (!generateAdr(args[0], out adr, pass)) return false;
                interp.memory[bwp++] = adr;
                interp.memory[bwp++] = 0xf000;	//SVC 1
                interp.memory[bwp++] = 0x0001;
                interp.memory[bwp++] = 0x1110;	//ST args[1],GR1
                if (!generateAdr(args[1], out adr, pass)) return false;
                interp.memory[bwp++] = adr;
                interp.memory[bwp++] = 0x7110;	//POP GR1
            }
            else if (mnem == "OUT")
            {
                interp.memory[bwp++] = 0x7001;	//PUSH GR1
                interp.memory[bwp++] = 0;
                interp.memory[bwp++] = 0x7002;	//PUSH GR2
                interp.memory[bwp++] = 0;
                interp.memory[bwp++] = 0x1210;	//LAD GR1,args[0]
                if (!generateAdr(args[0], out adr, pass)) return false;
                interp.memory[bwp++] = adr;
                interp.memory[bwp++] = 0x1020;	//LD GR2,args[1]
                if (!generateAdr(args[1], out adr, pass)) return false;
                interp.memory[bwp++] = adr;
                interp.memory[bwp++] = 0xf000;	//SVC 2
                interp.memory[bwp++] = 0x0002;
                interp.memory[bwp++] = 0x7120;	//POP GR2
                interp.memory[bwp++] = 0x7110;	//POP GR1
            }
            else if (mnem == "RPUSH")
                for (int i = 1; i <= 7; i++)
                {
                    interp.memory[bwp++] = (ushort)(0x7000 + i);	//PUSH GRn
                    interp.memory[bwp++] = 0;
                }
            else if (mnem == "RPOP")
                for (int i = 7; i >= 1; i--) interp.memory[bwp++] = (ushort)(0x7100 + (i << 4));	//POP GRn
            else if (args[2].Length > 0)	// has three arguments
            {
                if (!r_adr_x.TryGetValue(mnem, out inst)) return syntaxError(line);
                if (!generateR1(args[0], ref inst)) return syntaxError(line);
                if (!generateAdr(args[1], out adr, pass)) return false;
                if (!generateX(args[2], ref inst)) return syntaxError(line);
                interp.memory[bwp++] = inst;
                interp.memory[bwp++] = adr;
            }
            else if (args[1].Length > 0)	// has two arguments
                if (getRegisterNumber(args[0]) >= 0 && getRegisterNumber(args[1]) >= 0)
                {
                    if (!r1_r2.TryGetValue(mnem, out inst)) return syntaxError(line);
                    if (!generateR1(args[0], ref inst)) return syntaxError(line);
                    if (!generateR2(args[1], ref inst)) return syntaxError(line);
                    interp.memory[bwp++] = inst;
                }
                else
                {
                    if (adr_x.TryGetValue(mnem, out inst))
                    {
                        if (!generateAdr(args[0], out adr, pass)) return false;
                        if (!generateX(args[1], ref inst)) return syntaxError(line);
                        interp.memory[bwp++] = inst;
                        interp.memory[bwp++] = adr;
                    }
                    else
                    {
                        if (!r_adr_x.TryGetValue(mnem, out inst)) return syntaxError(line);
                        if (!generateR1(args[0], ref inst)) return syntaxError(line);
                        if (!generateAdr(args[1], out adr, pass)) return false;
                        interp.memory[bwp++] = inst;
                        interp.memory[bwp++] = adr;
                    }
                }
            else if (args[0].Length > 0)	// has one arguments
                if (getRegisterNumber(args[0]) >= 0)
                {
                    if (!r.TryGetValue(mnem, out inst)) return syntaxError(line);
                    if (!generateR1(args[0], ref inst)) return syntaxError(line);
                    interp.memory[bwp++] = inst;
                }
                else
                {
                    if (!adr_x.TryGetValue(mnem, out inst)) return syntaxError(line);
                    if (!generateAdr(args[0], out adr, pass)) return false;
                    interp.memory[bwp++] = inst;
                    interp.memory[bwp++] = adr;
                }
            else // has no argument
                if (!none.TryGetValue(mnem, out inst)) return syntaxError(line);
                else interp.memory[bwp++] = inst;

            return true;
        }

        // パス1回分のアセンブル
        // 引数passを1と2で2回呼ばれる
        public bool assemble(string srcFileName, int pass)
        {
            bool endFlag = false;

            if (interp.traceMode) Console.WriteLine("starting pass {0}", pass);

            using (var reader = new System.IO.StreamReader(srcFileName, Encoding.Default))
            {
                if (pass == 1) labelTable.Clear();
                interp.PR = 0;	// default program start address is 0
                bwp = 0;	// from zero address
                constReservePtr = 0xf000;	// constant reserve area from 0xf000
                for (; ; )
                {
                    string result = reader.ReadLine();
                    if (result == null) break;
                    if (!assembleLine(result, pass, ref endFlag)) return false;
                    if (endFlag) break;
                }
                return true;
            }
        }

        public Assembler(Interpreter interp)
        {
            this.interp = interp;
        }
    }
}
