﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.CSharp;
using System.CodeDom.Compiler;

namespace COMET2Lib
{
    // 便宜上のインタプリタ
    public class Interpreter
    {
        // COMET IIの世界
        public ushort[] memory = new ushort[65536];// メモリ世界
        public ushort PR;// レジスタ世界
        public bool[] validGotoAddress = new bool[65536];
        public int[] sourceLine = new int[65536];
    }

    public class Assembler
    {
        private Interpreter interp;

        // 5パターンに分けられた命令表
        private Dictionary<string, ushort> none = new Dictionary<string, ushort>()
        {
            {"NOP", 0x0 },
            {"RET", 0x8100 },
        };
        private Dictionary<string, ushort> r_adr_x = new Dictionary<string, ushort>()
        {
            { "LD", 0x1000 },
            { "ST", 0x1100 },
            { "LAD", 0x1200 },
            { "ADDA", 0x2000 },
            { "SUBA", 0x2100 },
            { "ADDL", 0x2200 },
            { "SUBL", 0x2300 },
            { "AND", 0x3000 },
            { "OR", 0x3100 },
            { "XOR", 0x3200 },
            { "CPA", 0x4000 },
            { "CPL", 0x4100 },
            { "SLA", 0x5000 },
            { "SRA", 0x5100 },
            { "SLL", 0x5200 },
            { "SRL", 0x5300 },
        };
        private Dictionary<string, ushort> r1_r2 = new Dictionary<string, ushort>()
        {
            { "LD", 0x1400 },
            { "ADDA", 0x2400 },
            { "SUBA", 0x2500 },
            { "ADDL", 0x2600 },
            { "SUBL", 0x2700 },
            { "AND", 0x3400 },
            { "OR", 0x3500 },
            { "XOR", 0x3600 },
            { "CPA", 0x4400 },
            { "CPL", 0x4500 },
        };
        private Dictionary<string, ushort> adr_x = new Dictionary<string, ushort>()
        {
            { "JMI", 0x6100 },
            { "JNZ", 0x6200 },
            { "JZE", 0x6300 },
            { "JUMP", 0x6400 },
            { "JPL", 0x6500 },
            { "JOV", 0x6600 },
            { "PUSH", 0x7000 },
            { "CALL", 0x8000 },
            { "SVC", 0xF000 },
        };
        private Dictionary<string, ushort> r = new Dictionary<string, ushort>()
        {
            { "POP", 0x7100 },
        };

        // アセンブル中のファイル名
        private string fileName;

        // アセンブル中の現在行
        private string currentLine;

        // アセンブル中の行番号
        private int lineNumber;

        // アセンブラが生成したデータを書いていくポインタ
        private ushort bwp;	// binary writing pointer

        // 構文エラーの通知
        private bool syntaxError(string srcLine)
        {
            Console.Error.WriteLine("Syntax Error: {0}", srcLine);
            return false;
        }

        // ホワイトスペースの判定
        private bool iswhitespace(char ch) { return ch == ' ' || ch == '\t'; }

        // 文字列のレジスタ名からレジスタ番号を整数で抜き出す
        private int getRegisterNumber(string str)
        {
            if (str.Length != 3) return -1;
            if (str[0] != 'G' || str[1] != 'R') return -1;
            if (str[2] < '0' || str[2] > '7') return -1;
            return str[2] - '0';
        }

        // 1行の中の1トークンを解析して取り出す
        private string[] divideTokens(string src)
        {
            var list = new List<string>();
            bool quoting = false, skipping = false;
            StringBuilder sb = new StringBuilder();
            foreach (char ch in src)
            {
                // いい加減な実装だが、それでも'…''…'を適切にさばける
                // ''→'の処理は別の場所でやるので、ここで意識する必要はない
                if (ch == '\'') quoting = !quoting;
                if (ch == ';') break;
                if (skipping)
                {
                    skipping = iswhitespace(ch);
                    if (skipping) continue;
                }
                else if (!quoting && (iswhitespace(ch) || ch == ','))
                {
                    list.Add(sb.ToString());
                    sb.Length = 0;
                    skipping = true;
                    continue;
                }
                sb.Append(ch);
            }
            list.Add(sb.ToString());
            return list.ToArray();
        }

        // ラベルテーブル
        private Dictionary<string, ushort> labelTable = new Dictionary<string, ushort>();

        // 引数のr1を命令第1語に反映
        private bool generateR1(string str, ref ushort inst)
        {
            int n = getRegisterNumber(str);
            inst |= (ushort)(n << 4);
            return n >= 0;
        }

        // 引数のr2を命令第1語に反映
        private bool generateR2(string str, ref ushort inst)
        {
            int n = getRegisterNumber(str);
            inst |= (ushort)n;
            return n >= 0;
        }

        // 引数のxを命令第1語に反映
        // generateR2とは正当なレジスタの判定条件が違うことに注意
        // R2にGR0はあるが、XにGR0は無い
        private bool generateX(string str, ref ushort inst)
        {
            int n = getRegisterNumber(str);
            inst |= (ushort)n;
            return n > 0;
        }

        // 定数を解析する
        private bool constantParser(string str, out ushort target, int pass)
        {
            if (ushort.TryParse(str, out target)) return true;
            if (str[0] == '#' && ushort.TryParse(str.Substring(1), System.Globalization.NumberStyles.AllowHexSpecifier, null, out target)) return true;
            if (pass == 1 || labelTable.TryGetValue(str, out target)) return true;
            Console.Error.WriteLine("label {0} not found", str);
            return false;
        }

        // =で前置された即値を格納する場所を示す
        private ushort constReservePtr;

        // adrの値を生成する
        // パス1では0を入れているところと、
        // =で前置された即値の処理が違うことに注意
        private bool generateAdr(string str, out ushort adr, int pass)
        {
            if (!str.StartsWith("=")) return constantParser(str, out adr, pass);
            ushort target;
            adr = 0;
            if (!constantParser(str.Substring(1), out target, pass)) return false;
            if (pass == 1) return true;
            interp.memory[constReservePtr] = target;
            adr = constReservePtr++;
            return true;
        }

        // DC命令専用の値の生成
        private bool generateDC(string str, int pass)
        {
            ushort target;
            if (str.Length == 0) return true;
            if (str[0] == '\'')
            {
                str = str.Substring(1);
                while (str.Length != 0)
                {
                    if (str[0] == '\'')
                    {
                        if (str.Length < 2 || str[1] != '\'') break;
                        str = str.Substring(1);
                    }
                    interp.memory[bwp++] = str[0];
                    str = str.Substring(1);
                }
            }
            else
                if (!constantParser(str, out target, pass)) return false;
                else interp.memory[bwp++] = target;
            return true;
        }

        // 1行のアセンブル
        private bool assembleLine(string line, int pass, ref bool endFlag)
        {
            ushort inst, adr;

            // 行の解析結果を入れる変数
            string[] tokens = divideTokens(line);
            string label = tokens.Length >= 1 ? tokens[0] : string.Empty;
            string mnem = tokens.Length >= 2 ? tokens[1] : string.Empty;
            string[] args = new string[3];
            for (int i = 0; i < 3; i++)
                args[i] = tokens.Length >= i + 3 ? tokens[i + 2] : string.Empty;

            if (pass == 1 && label.Length > 0)
            {
                if (labelTable.ContainsKey(label))
                {
                    Console.Error.WriteLine("label {0} already defined", label);
                    return false;
                }
                labelTable[label] = bwp;
            }
            if (mnem.Length == 0)
            { }	// empty mnem
            else if (mnem == "START")
                interp.PR = bwp;
            else if (mnem == "END")
                endFlag = true;
            else if (mnem == "DS")
                bwp += (ushort)int.Parse(args[0]);	// エラーチェック手抜き
            else if (mnem == "DC")
                for (int i = 2; i < tokens.Length; i++)
                    if (!generateDC(tokens[i], pass)) return false; else { }
            else if (mnem == "IN")
            {
                interp.validGotoAddress[bwp] = true;
                interp.memory[bwp++] = 0x7001;	//PUSH GR1
                interp.memory[bwp++] = 0;
                interp.validGotoAddress[bwp] = true;
                interp.memory[bwp++] = 0x1210;	//LAD GR1,args[0]
                if (!generateAdr(args[0], out adr, pass)) return false;
                interp.memory[bwp++] = adr;
                interp.validGotoAddress[bwp] = true;
                interp.memory[bwp++] = 0xf000;	//SVC 1
                interp.memory[bwp++] = 0x0001;
                interp.validGotoAddress[bwp] = true;
                interp.memory[bwp++] = 0x1110;	//ST args[1],GR1
                if (!generateAdr(args[1], out adr, pass)) return false;
                interp.memory[bwp++] = adr;
                interp.validGotoAddress[bwp] = true;
                interp.memory[bwp++] = 0x7110;	//POP GR1
            }
            else if (mnem == "OUT")
            {
                interp.validGotoAddress[bwp] = true;
                interp.memory[bwp++] = 0x7001;	//PUSH GR1
                interp.memory[bwp++] = 0;
                interp.validGotoAddress[bwp] = true;
                interp.memory[bwp++] = 0x7002;	//PUSH GR2
                interp.memory[bwp++] = 0;
                interp.validGotoAddress[bwp] = true;
                interp.memory[bwp++] = 0x1210;	//LAD GR1,args[0]
                if (!generateAdr(args[0], out adr, pass)) return false;
                interp.memory[bwp++] = adr;
                interp.validGotoAddress[bwp] = true;
                interp.memory[bwp++] = 0x1020;	//LD GR2,args[1]
                if (!generateAdr(args[1], out adr, pass)) return false;
                interp.memory[bwp++] = adr;
                interp.validGotoAddress[bwp] = true;
                interp.memory[bwp++] = 0xf000;	//SVC 2
                interp.memory[bwp++] = 0x0002;
                interp.validGotoAddress[bwp] = true;
                interp.memory[bwp++] = 0x7120;	//POP GR2
                interp.validGotoAddress[bwp] = true;
                interp.memory[bwp++] = 0x7110;	//POP GR1
            }
            else if (mnem == "RPUSH")
                for (int i = 1; i <= 7; i++)
                {
                    interp.validGotoAddress[bwp] = true;
                    interp.memory[bwp++] = (ushort)(0x7000 + i);	//PUSH GRn
                    interp.memory[bwp++] = 0;
                }
            else if (mnem == "RPOP")
                for (int i = 7; i >= 1; i--)
                {
                    interp.validGotoAddress[bwp] = true;
                    interp.memory[bwp++] = (ushort)(0x7100 + (i << 4));	//POP GRn
                }
            else if (args[2].Length > 0)	// has three arguments
            {
                interp.validGotoAddress[bwp] = true;
                if (!r_adr_x.TryGetValue(mnem, out inst)) return syntaxError(line);
                if (!generateR1(args[0], ref inst)) return syntaxError(line);
                if (!generateAdr(args[1], out adr, pass)) return false;
                if (!generateX(args[2], ref inst)) return syntaxError(line);
                interp.memory[bwp++] = inst;
                interp.memory[bwp++] = adr;
            }
            else if (args[1].Length > 0)	// has two arguments
            {
                interp.validGotoAddress[bwp] = true;
                if (getRegisterNumber(args[0]) >= 0 && getRegisterNumber(args[1]) >= 0)
                {
                    if (!r1_r2.TryGetValue(mnem, out inst)) return syntaxError(line);
                    if (!generateR1(args[0], ref inst)) return syntaxError(line);
                    if (!generateR2(args[1], ref inst)) return syntaxError(line);
                    interp.memory[bwp++] = inst;
                }
                else
                {
                    if (adr_x.TryGetValue(mnem, out inst))
                    {
                        if (!generateAdr(args[0], out adr, pass)) return false;
                        if (!generateX(args[1], ref inst)) return syntaxError(line);
                        interp.memory[bwp++] = inst;
                        interp.memory[bwp++] = adr;
                    }
                    else
                    {
                        if (!r_adr_x.TryGetValue(mnem, out inst)) return syntaxError(line);
                        if (!generateR1(args[0], ref inst)) return syntaxError(line);
                        if (!generateAdr(args[1], out adr, pass)) return false;
                        interp.memory[bwp++] = inst;
                        interp.memory[bwp++] = adr;
                    }
                }
            }
            else if (args[0].Length > 0)	// has one arguments
            {
                interp.validGotoAddress[bwp] = true;
                if (getRegisterNumber(args[0]) >= 0)
                {
                    if (!r.TryGetValue(mnem, out inst)) return syntaxError(line);
                    if (!generateR1(args[0], ref inst)) return syntaxError(line);
                    interp.memory[bwp++] = inst;
                }
                else
                {
                    if (!adr_x.TryGetValue(mnem, out inst)) return syntaxError(line);
                    if (!generateAdr(args[0], out adr, pass)) return false;
                    interp.memory[bwp++] = inst;
                    interp.memory[bwp++] = adr;
                }
            }
            else // has no argument
            {
                interp.validGotoAddress[bwp] = true;
                if (!none.TryGetValue(mnem, out inst)) return syntaxError(line);
                else interp.memory[bwp++] = inst;
            }
            return true;
        }

        public static bool CompileCSharpCode(String sourceFile, String exeFile)
        {
            // C#コンパイラを用意する
            CSharpCodeProvider provider = new CSharpCodeProvider();

            // コンパイラの引数を用意する
            CompilerParameters cp = new CompilerParameters();

            // 参照するライブラリを指定する
            cp.ReferencedAssemblies.Add("System.dll");

            // 実行ファイルを作成する
            cp.GenerateExecutable = true;
            cp.OutputAssembly = exeFile;

            // メモリではなくファイルに作成する
            cp.GenerateInMemory = false;

            // デバッグビルドにする場合は有効にする
            //cp.IncludeDebugInformation = true;

            // コンパイル実行
            CompilerResults cr = provider.CompileAssemblyFromFile(cp, sourceFile);
            if (cr.Errors.Count == 0) return true;
            // コンパイルエラー
            Console.Error.WriteLine("{0}から{1}を作成する時にエラーが発生しました。",
                sourceFile, cr.PathToAssembly);
            foreach (CompilerError ce in cr.Errors)
            {
                Console.Error.WriteLine("  {0}", ce.ToString());
                Console.Error.WriteLine();
            }
            return false;
        }

        // パス1回分のアセンブル
        // 引数passを1と2で2回呼ばれる
        public bool assemble(string srcFileName, int pass)
        {
            this.fileName = srcFileName;
            bool endFlag = false;

            using (var reader = new System.IO.StreamReader(srcFileName, Encoding.Default))
            {
                if (pass == 1)
                {
                    labelTable.Clear();
                }
                else
                {
                    // constant reserve area from last address of pass1
                    constReservePtr = bwp;
                }
                interp.PR = 0;	// default program start address is 0
                bwp = 0;	// from zero address
                lineNumber = 1;
                for (; ; )
                {
                    interp.sourceLine[bwp] = lineNumber;
                    currentLine = reader.ReadLine();
                    if (currentLine == null) break;
                    if (!assembleLine(currentLine, pass, ref endFlag)) return false;
                    if (endFlag) break;
                    lineNumber++;
                }
            }
            return true;
        }

        private string createIntemeditateLabelName(int number)
        {
            return string.Format("l{0:x4}", number);
        }

        private void createCodes(System.IO.StreamWriter writer)
        {
            ushort srcp = 0;
            // この条件判断は本当は危うい。bwpが0xffffを越えて0になったときに誤判定する
            // しかし、そのような事態があればメインからRETで戻って終了も出来なくなるので
            // あり得ない異常事態と見なしてそのまま放置する
            while (srcp < bwp)
            {
                if (!interp.validGotoAddress[srcp]) { srcp++; continue; }

                // ラベル
                writer.Write("{0}: ", createIntemeditateLabelName(srcp));

                // 解析側変数宣言
                ushort firstOP = interp.memory[srcp++];
                ushort OP = (ushort)(firstOP >> 8);
                ushort mainOP = (ushort)(OP >> 4);
                ushort subOP = (ushort)(OP & 0xf);
                ushort r1 = (ushort)((firstOP >> 4) & 0xf);
                ushort r2 = (ushort)((firstOP & 0xf));
                string value = "";

                // 実行側変数初期化とvalue解析
                writer.Write("PR++;");
                if ((mainOP >= 1 && mainOP <= 5 && subOP <= 3)
                    || mainOP == 6 || OP == 0x70 || OP == 0x80 || OP == 0xf0)
                {
                    if (r2 == 0)
                    {
                        value = "(memory[PR-1])";
                    }
                    else
                    {
                        value = "((ushort)(GR[" + r2 + "] + memory[PR-1]))";
                    }
                    writer.Write("PR++;");
                    srcp++;
                    if (OP == 0x10 || (mainOP >= 2 && mainOP <= 5))
                        value = "(memory[" + value + "])";
                }
                else if (mainOP >= 1 && mainOP <= 5 && subOP >= 4)
                {
                    if (r2 == 0)
                    {
                        value = "((ushort)0)";
                    }
                    else
                    {
                        value = "(GR[" + r2 + "])";
                    }
                }
                // 分類2
                switch (OP & 0xfb)
                {
                    case 0x10:    // LD
                        writer.Write("GR[" + r1 + "] = setFR(" + value + ", false);");
                        break;
                    case 0x20:	// ADDA
                        writer.Write("GR[" + r1 + "] = setFRbyA((short)GR[" + r1 + "] + (short)" + value + ");");
                        break;
                    case 0x21:	// SUBA
                        writer.Write("GR[" + r1 + "] = setFRbyA((short)GR[" + r1 + "] - (short)" + value + ");");
                        break;
                    case 0x22:	// ADDL
                        writer.Write("GR[" + r1 + "] = setFRbyL(GR[" + r1 + "] + " + value + ");");
                        break;
                    case 0x23:	// SUBL
                        writer.Write("GR[" + r1 + "] = setFRbyL(GR[" + r1 + "] - " + value + ");");
                        break;
                    case 0x30:	// AND
                        writer.Write("GR[" + r1 + "] = setFR((ushort)(GR[" + r1 + "] & " + value + "), false);");
                        break;
                    case 0x31:	// OR
                        writer.Write("GR[" + r1 + "] = setFR((ushort)(GR[" + r1 + "] | " + value + "), false);");
                        break;
                    case 0x32:	// XOR
                        writer.Write("GR[" + r1 + "] = setFR((ushort)(GR[" + r1 + "] ^ " + value + "), false);");
                        break;
                    case 0x40:	// CPA
                        writer.Write("FR = (ushort)((GR[" + r1 + "] == " + value + " ? 1 : 0) | ((short)GR[" + r1 + "] < (short)" + value + " ? 2 : 0));");
                        break;
                    case 0x41:	// CPL
                        writer.Write("FR = (ushort)((GR[" + r1 + "] == " + value + " ? 1 : 0) | (GR[" + r1 + "] < " + value + " ? 2 : 0));");
                        break;
                    case 0x50:	// SLA
                        writer.Write("GR[" + r1 + "] = setFR((ushort)(((GR[" + r1 + "] << " + value + ") & 0x7fff) | (GR[" + r1 + "] & 0x8000)), ((GR[" + r1 + "] << " + value + ") & 0x8000) != 0);");
                        break;
                    case 0x51:	// SRA
                        // 処理系によっては意図しない結果になる可能性がある
                        writer.Write("GR[" + r1 + "] = setFR((ushort)((short)GR[" + r1 + "] >> " + value + "), " + value + " > 0 && ((GR[" + r1 + "] >> (" + value + " - 1)) & 1) != 0);");
                        break;
                    case 0x52:	// SLL
                        writer.Write("GR[" + r1 + "] = setFR((ushort)(GR[" + r1 + "] << " + value + "), " + value + " > 0 && ((GR[" + r1 + "] << (" + value + " - 1)) & 0x8000) != 0);");
                        break;
                    case 0x53:	// SRL
                        writer.Write("GR[" + r1 + "] = setFR((ushort)(GR[" + r1 + "] >> " + value + "), " + value + " > 0 && ((GR[" + r1 + "] >> (" + value + " - 1)) & 1) != 0);");
                        break;
                    default:
                        switch (OP)
                        {
                            case 0x00:	// NOP
                                break;
                            case 0x11:    // ST r,adr,x store
                                writer.Write("memory[" + value + "] = GR[" + r1 + "];");
                                break;
                            case 0x12:    // LAD r,adr,x load address
                                writer.Write("GR[" + r1 + "] = " + value + ";");
                                break;
                            case 0x61:	// JMI
                                writer.Write("if ((FR & (ushort)FlagMasks.SF) != 0) PR = " + value + ";");
                                writer.Write("goto ongoto;");
                                break;
                            case 0x62:	// JNZ
                                writer.Write("if ((FR & (ushort)FlagMasks.ZF) == 0) PR = " + value + ";");
                                writer.Write("goto ongoto;");
                                break;
                            case 0x63:	// JZE
                                writer.Write("if ((FR & (ushort)FlagMasks.ZF) != 0) PR = " + value + ";");
                                writer.Write("goto ongoto;");
                                break;
                            case 0x64:	// JUMP
                                writer.Write("PR = " + value + ";");
                                writer.Write("goto ongoto;");
                                break;
                            case 0x65:	// JPL
                                writer.Write("if ((FR & ((ushort)FlagMasks.SF | (ushort)FlagMasks.ZF)) == 0) PR = " + value + ";");
                                writer.Write("goto ongoto;");
                                break;
                            case 0x66:	// JOV
                                writer.Write("if ((FR & (ushort)FlagMasks.OF) != 0) PR = " + value + ";");
                                writer.Write("goto ongoto;");
                                break;
                            case 0x70:	// PUSH
                                writer.Write("memory[--SP] = " + value + ";");
                                break;
                            case 0x71:	// POP
                                writer.Write("GR[" + r1 + "] = memory[SP++];");
                                break;
                            case 0x80:	// CALL
                                writer.Write("memory[--SP] = PR;");
                                writer.Write("PR = " + value + ";");
                                writer.Write("goto ongoto;");
                                break;
                            case 0x81:	// RET
                                writer.Write("PR = memory[SP++];");
                                writer.Write("goto ongoto;");
                                break;
                            case 0xf0:	// SVC
                                writer.Write("if( svc(" + value + ") ) return;");
                                break;
                        }
                        break;
                }
                writer.WriteLine();
            }
        }

        public bool CreateIntermeditateSource(string outputAssembly)
        {
            string intermeditateSrcFileName = System.IO.Path.ChangeExtension(outputAssembly, ".cs");
            using (var writer = new System.IO.StreamWriter(intermeditateSrcFileName, false, Encoding.Default))
            {
                writer.Write(@"using System;
using System.Collections.Generic;
using System.Text;

class Program
{
    private enum FlagMasks { OF = 4, SF = 2, ZF = 1 };

    private static ushort[] GR = new ushort[8];
    private static ushort PR=0, SP=0xfffe, FR=0;
    private static ushort[] memory = new ushort[65536];
    private static ushort firstOP,OP,mainOP,subOP,r1,r2,value;

    // スーパバイザコールのINPUT処理
    private static void svcInput()
    {
        string buffer = Console.ReadLine();
        if (buffer == null)
        {
            GR[2] = 0xffff;
            return;
        }
        GR[2] = (ushort)buffer.Length;
        for (int i = 0; i < buffer.Length; i++)
        {
            memory[GR[1] + i] = buffer[i];
        }
    }

    // スーパバイザコールのOUTPUT処理
    private static void svcOutput()
    {
        int i;
        for (i = 0; i < GR[2]; i++)
            Console.Write(""{0}"", (char)(memory[GR[1] + i] & 0xff));
        Console.Write('\n');
    }

    // スーパバイザコール
    private static bool svc( ushort value )
    {
        switch (value)
        {
            case 0:	return true; // EXIT
            case 1:	svcInput(); break; // IN
            case 2:	svcOutput(); break; // OUT
        }
        return false;
    }

    // フラグレジスタの設定
    private static ushort setFR(ushort result, bool OFvalue)
    {
        FR = (ushort)(((result == 0) ? (int)FlagMasks.ZF : 0) | ((result & 0x8000) >> 14) | (OFvalue ? (int)FlagMasks.OF : 0));
        return result;
    }

    // 算術命令(……A)の場合のフラグレジスタの設定
    private static ushort setFRbyA(int result)
    {
        return setFR((ushort)result, result < -32768 || result > 32767);
    }

    // 論理命令(……L)の場合のフラグレジスタの設定
    private static ushort setFRbyL(int result)
    {
        return setFR((ushort)result, result < 0 || result > 65535);
    }

    static void Main(string[] args)
    {
        memory[0xfffe] = 0xf000;	// SVC 0=停止を書き込んでおく
        memory[0xffff] = 0x0000;
        memory[--SP] = 0xfffe;	//停止命令のあるアドレスをPUSHしておく
        ushort[] initialImage = 
        {
");
                // メモリ初期化コードの生成
                for (ushort i = 0; i < constReservePtr; i++)
                {
                    writer.Write("0x{0:x4},", interp.memory[i]);
                    if ((i % 8) == 7) writer.WriteLine();
                }
                writer.Write(@"
        };
        for (int i = 0; i < initialImage.Length; i++) memory[i] = initialImage[i];
");

                // 実行コード本体の生成
                createCodes(writer);

                // ongotoの生成
                writer.Write(@"
ongoto:
switch(PR)
{
");
                for (ushort i = 0; i < bwp; i++)
                    if (interp.validGotoAddress[i])
                        writer.WriteLine("case 0x{0:x4}: goto {1};", i, createIntemeditateLabelName(i));

                // 決め打ちアドレス対応コードを書く
                // lfffeはメインからRETしたときに飛ぶ先
                // デフォルトはシステム例外に対応
                writer.Write(@"
case 0xfffe:
    return;
default:
         Console.WriteLine(""異常終了します。理由: 命令が記述されていないアドレス[{0}]への移動がリクエストされました。"",PR.ToString(""x4""));
    return;
}
");
                writer.Write(@"
    }
}
");
            }
            CompileCSharpCode(intermeditateSrcFileName, outputAssembly);
            return true;
        }

        public Assembler(Interpreter interp)
        {
            this.interp = interp;
        }
    }
}
